"""An opionionated wrapper around :class:`pyg4ometry.visualization.VtkViewerNew`."""

from __future__ import annotations

import argparse
import logging
from pathlib import Path

import pyg4ometry.geant4 as g4
import vtk
from pyg4ometry import config as meshconfig
from pyg4ometry import gdml
from pyg4ometry import visualisation as pyg4vis

from .utils import load_dict
from .visualization import load_color_auxvals_recursive

log = logging.getLogger(__name__)


def visualize(registry: g4.Registry, scenes: dict | None = None, points=None) -> None:
    if scenes is None:
        scenes = {}

    v = pyg4vis.VtkViewerColouredNew()
    v.addLogicalVolume(registry.worldVolume)

    load_color_auxvals_recursive(registry.worldVolume)
    registry.worldVolume.pygeom_color_rgba = False  # hide the wireframe of the world.
    _color_recursive(registry.worldVolume, v, scenes.get("color_overrides", {}))

    for clip in scenes.get("clipper", []):
        v.addClipper(clip["origin"], clip["normal"], bClipperCloseCuts=False)

    v.buildPipelinesAppend()
    v.addAxes(length=5000)
    v.axes[0].SetVisibility(False)  # hide axes by default.

    if points is not None:
        _add_points(v, points)

    # override the interactor style.
    v.interactorStyle = _KeyboardInteractor(v.ren, v.iren, v, scenes)
    v.interactorStyle.SetDefaultRenderer(v.ren)
    v.iren.SetInteractorStyle(v.interactorStyle)

    # set some defaults
    if "default" in scenes:
        sc = scenes["default"]
        _set_camera(
            v,
            up=sc.get("up"),
            pos=sc.get("camera"),
            focus=sc.get("focus"),
            parallel=sc.get("parallel", False),
        )
    else:
        _set_camera(v, up=(1, 0, 0), pos=(0, 0, +20000))

    v.view()


class _KeyboardInteractor(vtk.vtkInteractorStyleTrackballCamera):
    def __init__(self, renderer, iren, vtkviewer, scenes):
        self.AddObserver("KeyPressEvent", self.keypress)

        self.ren = renderer
        self.iren = iren
        self.vtkviewer = vtkviewer
        self.scenes = scenes

    def keypress(self, _obj, _event):
        # predefined: _e_xit

        key = self.iren.GetKeySym()
        if key == "a":  # toggle _a_xes
            ax = self.vtkviewer.axes[0]
            ax.SetVisibility(not ax.GetVisibility())
            self.ren.GetRenderWindow().Render()

        if key == "v" and self.vtkviewer.points is not None:  # toggle _v_ertices
            pn = self.vtkviewer.points
            pn.SetVisibility(not ax.GetVisibility())
            self.ren.GetRenderWindow().Render()

        if key == "u":  # _u_p
            _set_camera(self.vtkviewer, up=(0, 0, 1), pos=(-20000, 0, 0))

        if key == "t":  # _t_op
            _set_camera(self.vtkviewer, up=(1, 0, 0), pos=+20000)

        if key == "p":  # _p_arralel projection
            cam = self.ren.GetActiveCamera()
            _set_camera(self.vtkviewer, parallel=not cam.GetParallelProjection())

        sc_index = 1
        for sc in self.scenes.get("scenes", []):
            if key == f"F{sc_index}":
                _set_camera(
                    self.vtkviewer,
                    up=sc.get("up"),
                    pos=sc.get("camera"),
                    focus=sc.get("focus"),
                    parallel=sc.get("parallel", False),
                )
                sc_index += 1

        if key == "s":  # _s_ave
            _export_png(self.vtkviewer)

        if key == "i":  # dump camera _i_nfo
            cam = self.ren.GetActiveCamera()
            print(f"- focus: {list(cam.GetFocalPoint())}")  # noqa: T201
            print(f"  up: {list(cam.GetViewUp())}")  # noqa: T201
            print(f"  camera: {list(cam.GetPosition())}")  # noqa: T201
            if cam.GetParallelProjection():
                print(f"  parallel: {cam.GetParallelScale()}")  # noqa: T201

        if key == "plus":
            _set_camera(self.vtkviewer, dolly=1.1)
        if key == "minus":
            _set_camera(self.vtkviewer, dolly=0.9)


def _set_camera(
    v: pyg4vis.VtkViewerColouredNew,
    focus: tuple[float, float, float] | None = None,
    up: tuple[float, float, float] | None = None,
    pos: tuple[float, float, float] | None = None,
    dolly: float | None = None,
    parallel: bool | int | None = None,
) -> None:
    cam = v.ren.GetActiveCamera()
    if focus is not None:
        cam.SetFocalPoint(*focus)
    if up is not None:
        cam.SetViewUp(*up)
    if pos is not None:
        cam.SetPosition(*pos)
    if dolly is not None:
        if cam.GetParallelProjection():
            cam.SetParallelScale(1 / dolly * cam.GetParallelScale())
        else:
            cam.Dolly(dolly)
    if parallel is not None:
        cam.SetParallelProjection(int(parallel) > 0)
        if cam.GetParallelScale() == 1.0:
            # still at initial value, set to something more useful.
            cam.SetParallelScale(2000)
        if not isinstance(parallel, bool):
            cam.SetParallelScale(int(parallel))

    v.ren.ResetCameraClippingRange()
    v.ren.GetRenderWindow().Render()


def _export_png(v: pyg4vis.VtkViewerColouredNew, file_name="scene.png") -> None:
    ifil = vtk.vtkWindowToImageFilter()
    ifil.SetInput(v.renWin)
    ifil.ReadFrontBufferOff()
    ifil.Update()

    # get a non-colliding file name.
    p = Path(file_name)
    stem = p.stem
    idx = 0
    while p.exists():
        p = p.with_stem(f"{stem}_{idx}")
        idx += 1
        if idx > 1000:
            msg = "could not find file name"
            raise ValueError(msg)

    png = vtk.vtkPNGWriter()
    png.SetFileName(str(p.absolute()))
    png.SetInputConnection(ifil.GetOutputPort())
    png.Write()


def _add_points(v, points) -> None:
    # create vtkPolyData from points.
    vp = vtk.vtkPoints()
    ca = vtk.vtkCellArray()
    pd = vtk.vtkPolyData()

    for t in points:
        p = vp.InsertNextPoint(*t)
        ca.InsertNextCell(1)
        ca.InsertCellPoint(p)

    pd.SetPoints(vp)
    pd.SetVerts(ca)

    # add points to renderer.
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(pd)
    mapper.ScalarVisibilityOff()

    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetColor(1, 1, 0)
    actor.GetProperty().SetPointSize(5)
    actor.GetProperty().SetOpacity(1)
    actor.GetProperty().SetRenderPointsAsSpheres(True)

    v.ren.AddActor(actor)
    v.points = actor


def _color_recursive(
    lv: g4.LogicalVolume, viewer: pyg4vis.ViewerBase, overrides: dict
) -> None:
    if hasattr(lv, "pygeom_color_rgba") or lv.name in overrides:
        color_rgba = overrides.get(lv.name, lv.pygeom_color_rgba)
        for vis in viewer.instanceVisOptions[lv.name]:
            if color_rgba is False:
                vis.alpha = 0
                vis.visible = False
            else:
                vis.colour = color_rgba[0:3]
                vis.alpha = color_rgba[3]
                vis.visible = vis.alpha > 0

    for pv in lv.daughterVolumes:
        if pv.type == "placement":
            _color_recursive(pv.logicalVolume, viewer, overrides)


def vis_gdml_cli() -> None:
    parser = argparse.ArgumentParser(
        prog="legend-pygeom-vis",
        description="%(prog)s command line interface",
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="""Increase the program verbosity to maximum""",
    )
    parser.add_argument(
        "--fine",
        action="store_true",
        help="""use finer meshing settings""",
    )
    parser.add_argument(
        "--scene",
        "-s",
        help="""scene definition file.""",
    )
    parser.add_argument(
        "--add-points",
        help="""load points from LH5 file""",
    )
    parser.add_argument(
        "--add-points-columns",
        default="stp/vertices:xloc,yloc,zloc",
        help="""columns in the point file %(default)s""",
    )

    parser.add_argument(
        "filename",
        help="""GDML file to visualize.""",
    )

    args = parser.parse_args()

    if args.verbose:
        logging.getLogger("pygeomtools").setLevel(logging.DEBUG)
    if args.debug:
        logging.root.setLevel(logging.DEBUG)

    scene = {}
    if args.scene:
        scene = load_dict(args.scene)

    if scene.get("fine_mesh", args.fine):
        meshconfig.setGlobalMeshSliceAndStack(100)

    points = None
    if args.add_points:
        import pint
        from lgdo import lh5

        table_parts = [c.strip() for c in args.add_points_columns.split(":")]
        point_table = table_parts[0]
        point_columns = [c.strip() for c in table_parts[1].split(",")]
        if len(table_parts) != 2 or len(point_columns) != 3:
            msg = "invalid parameter for points"
            raise ValueError(msg)

        log.info(
            "loading table %s (with columns %s) from file %s",
            point_table,
            str(point_columns),
            args.add_points,
        )
        point_table = lh5.read(point_table, args.add_points)

        # the points need to be in mm.
        u = pint.get_application_registry()
        units = [u(point_table[c].getattrs().get("units", "")) for c in point_columns]
        units = [(un / u.mm).to("dimensionless").m for un in units]

        points = point_table.view_as("pd")[point_columns].to_numpy() * units

    log.info("loading GDML geometry from %s", args.filename)
    registry = gdml.Reader(args.filename).getRegistry()

    log.info("visualizing...")
    visualize(registry, scene, points)



from __future__ import annotations

import importlib.metadata

import pyg4ometry
from legendmeta import AttrsDict
from pyg4ometry import geant4 as g4

import pygeomtools


def test_package():
    assert importlib.metadata.version("legend-pygeom-tools") == pygeomtools.__version__


def test_detector_info(tmp_path):
    from pygeomtools import RemageDetectorInfo, detectors, geometry, visualization

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint1 = g4.LogicalVolume(
        scint, g4.MaterialPredefined("G4_lAr"), "scint1", registry
    )
    scint2 = g4.LogicalVolume(
        scint, g4.MaterialPredefined("G4_lAr"), "scint2", registry
    )
    scint1pv = g4.PhysicalVolume(
        [0, 0, 0], [-255, 0, 0], scint1, "scint1", world_lv, registry
    )
    scint1pv.pygeom_active_detector = RemageDetectorInfo("scintillator", 3)
    scint2pv = g4.PhysicalVolume(
        [0, 0, 0], [+255, 0, 0], scint2, "scint2", world_lv, registry
    )
    scint2pv.pygeom_active_detector = RemageDetectorInfo("scintillator", 3)

    det = g4.solid.Box("det", 0.1, 0.5, 0.5, registry, "m")
    det = g4.LogicalVolume(det, g4.MaterialPredefined("G4_Ge"), "det", registry)
    det1 = g4.PhysicalVolume([0, 0, 0], [0, 0, 0], det, "det1", scint1, registry)
    det1.pygeom_active_detector = RemageDetectorInfo("optical", 1, {"some": "metadata"})
    det2 = g4.PhysicalVolume([0, 0, 0], [0, 0, 0], det, "det2", scint2, registry)
    det2.pygeom_active_detector = RemageDetectorInfo(
        "germanium", 2, {"other": "other metadata"}
    )

    detectors.write_detector_auxvals(registry)
    visualization.write_color_auxvals(registry)
    geometry.check_registry_sanity(registry, registry)

    w = pyg4ometry.gdml.Writer()
    w.addDetector(registry)

    w.write(tmp_path / "geometry.gdml")

    # test read again
    registry = pyg4ometry.gdml.Reader(tmp_path / "geometry.gdml").getRegistry()

    assert detectors.get_sensvol_metadata(registry, "det2") == {
        "other": "other metadata"
    }
    det1meta = detectors.get_sensvol_metadata(registry, "det1")
    assert det1meta == {"some": "metadata"}
    assert det1meta.some == "metadata"
    assert isinstance(det1meta, AttrsDict)
    assert detectors.get_sensvol_metadata(registry, "scint1") is None
    sensvols = detectors.get_all_sensvols(registry)
    assert set(sensvols.keys()) == {"det2", "det1", "scint1", "scint2"}
    assert sensvols["scint1"].uid == 3



"""Assignment of sensitive detectors to physical volumes, for use in ``remage``."""

from __future__ import annotations

import json
import logging
from collections.abc import Generator
from dataclasses import dataclass
from itertools import groupby
from pathlib import Path
from typing import Literal

import pyg4ometry.geant4 as g4
from legendmeta import AttrsDict
from pyg4ometry.gdml.Defines import Auxiliary

log = logging.getLogger(__name__)


@dataclass
class RemageDetectorInfo:
    detector_type: Literal["optical", "germanium", "scintillator"]
    """``remage`` detector type."""

    uid: int
    """``remage`` detector UID."""

    metadata: object | None = None
    """Attach arbitrary metadata to this sensitive volume. This will be written to GDML as JSON.

    See also
    ========
    .get_sensvol_metadata
    """


def walk_detectors(
    pv: g4.PhysicalVolume | g4.LogicalVolume | g4.Registry,
) -> Generator[tuple[g4.PhysicalVolume, RemageDetectorInfo], None, None]:
    """Iterate over all physical volumes that have a :class:`RemageDetectorInfo` attached."""

    if isinstance(pv, g4.PhysicalVolume):
        det = None
        if hasattr(pv, "pygeom_active_detector"):
            det = pv.pygeom_active_detector
        elif hasattr(pv, "pygeom_active_dector"):
            import warnings

            warnings.warn(
                "pygeom_active_dector (typo!) is deprecated, use pygeom_active_detector instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            det = pv.pygeom_active_dector
        if det is not None:
            assert isinstance(det, RemageDetectorInfo)
            yield pv, det

    if isinstance(pv, g4.LogicalVolume):
        next_v = pv
    if isinstance(pv, g4.PhysicalVolume):
        next_v = pv.logicalVolume
    elif isinstance(pv, g4.Registry):
        next_v = pv.worldVolume
    else:
        msg = "invalid type encountered in walk_detectors volume tree"
        raise TypeError(msg)

    for dv in next_v.daughterVolumes:
        if dv.type == "placement":
            yield from walk_detectors(dv)


def generate_detector_macro(registry: g4.Registry, filename: str) -> None:
    """Create a Geant4 macro file containing the defined active detector volumes for use in remage."""

    macro_lines = {}

    for pv, det in walk_detectors(registry):
        if pv.name in macro_lines:
            return
        mac = f"/RMG/Geometry/RegisterDetector {det.detector_type.title()} {pv.name} {det.uid}\n"
        macro_lines[pv.name] = mac

    macro_contents = "".join(macro_lines.values())

    with Path(filename).open("w", encoding="utf-8") as f:
        f.write(macro_contents)


def write_detector_auxvals(registry: g4.Registry) -> None:
    """Append an auxiliary structure, storing the sensitive detector volume information.

    .. note::
        see <metadata> for a reference
    """
    written_pvs = set()
    group_it = groupby(walk_detectors(registry), lambda d: d[1].detector_type)

    meta_group_aux = Auxiliary("RMG_detector_meta", "", registry)

    for key, group in group_it:
        group_aux = Auxiliary("RMG_detector", key, registry)

        for pv, det in group:
            if pv.name in written_pvs:
                return
            written_pvs.add(pv.name)

            group_aux.addSubAuxiliary(
                Auxiliary(pv.name, det.uid, registry, addRegistry=False)
            )
            if det.metadata is not None:
                json_meta = json.dumps(det.metadata)
                meta_group_aux.addSubAuxiliary(
                    Auxiliary(pv.name, json_meta, registry, addRegistry=False)
                )


def get_sensvol_metadata(registry: g4.Registry, name: str) -> AttrsDict | None:
    """Load metadata attached to the given sensitive volume."""
    auxs = [aux for aux in registry.userInfo if aux.auxtype == "RMG_detector_meta"]
    if auxs == []:
        return None
    meta_aux = auxs[0]
    assert len(auxs) == 1

    meta_auxs = [aux for aux in meta_aux.subaux if aux.auxtype == name]
    if meta_auxs == []:
        return None
    assert len(meta_auxs) == 1
    return AttrsDict(json.loads(meta_auxs[0].auxvalue))


def get_all_sensvols(registry: g4.Registry) -> dict[str, RemageDetectorInfo]:
    """Load all registered sensitive detectors with their metadata."""
    auxs = [aux for aux in registry.userInfo if aux.auxtype == "RMG_detector_meta"]
    if auxs == []:
        meta_auxs = {}
    else:
        assert len(auxs) == 1
        meta_auxs = {
            aux.auxtype: AttrsDict(json.loads(aux.auxvalue)) for aux in auxs[0].subaux
        }

    detmapping = {}
    type_auxs = [aux for aux in registry.userInfo if aux.auxtype == "RMG_detector"]
    for type_aux in type_auxs:
        for det_aux in type_aux.subaux:
            detmapping[det_aux.auxtype] = RemageDetectorInfo(
                type_aux.auxvalue, int(det_aux.auxvalue), meta_auxs.get(det_aux.auxtype)
            )

    if set(meta_auxs.keys()) - set(detmapping.keys()) != set():
        msg = "invalid GDML auxval structure"
        raise RuntimeError(msg)

    return detmapping



from __future__ import annotations

from . import detectors, geometry, utils
from ._version import version as __version__
from .detectors import RemageDetectorInfo, get_all_sensvols, get_sensvol_metadata

# note: do not load viewer module here, as it has quite large nested imports.

__all__ = [
    "RemageDetectorInfo",
    "__version__",
    "detectors",
    "geometry",
    "get_all_sensvols",
    "get_sensvol_metadata",
    "utils",
]



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Callable

import yaml
from legendmeta import AttrsDict

log = logging.getLogger(__name__)

__file_extensions__ = {"json": [".json"], "yaml": [".yaml", ".yml"]}


def load_dict(fname: str, ftype: str | None = None) -> dict:
    """Load a text file as a Python dict."""
    fname = Path(fname)

    # determine file type from extension
    if ftype is None:
        for _ftype, exts in __file_extensions__.items():
            if fname.suffix in exts:
                ftype = _ftype

    msg = f"loading {ftype} dict from: {fname}"
    log.debug(msg)

    with fname.open() as f:
        if ftype == "json":
            return json.load(f)
        if ftype == "yaml":
            return yaml.safe_load(f)

        msg = f"unsupported file format {ftype}"
        raise NotImplementedError(msg)


def load_dict_from_config(
    config: dict, key: str, default: Callable[[], AttrsDict]
) -> AttrsDict:
    """Helper functions to load nested data from a config file.

    * If ``key`` is in the config file
      - and it refers to a string: load a JSON/YAML file from that path.
      - and it refers to a dict: use that directly
    * else, the default value is loaded via the ``default`` callable.
    """
    m = config.get(key)
    if isinstance(m, str):
        return AttrsDict(load_dict(m))
    if isinstance(m, dict):
        return AttrsDict(m)
    return default()
